#include <sys/ioctl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <string.h>
#include <ifret.h>

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif

struct ifret
{
   char ifname[IFNAMSIZ];
   char *ifdata;
   char *ifcur;
   int iflen;
};

/* interface retrieveal mechanism : should not be externally used */

/* initialize the interface retrieval mechanism */
ifret_t *
ifret_init(void)
{
   int sd;
   struct ifconf ifc;
   ifret_t *ifret;

   ifret = malloc(sizeof(ifret_t));
   if (!ifret)
      return (NULL);

   memset(ifret, 0, sizeof(*ifret));

#define IFDATSIZ (50 * sizeof(struct ifreq))

   ifret->ifdata = malloc(IFDATSIZ);
   if (!ifret->ifdata) {
      free(ifret);
      return (NULL);
   }
   memset(ifret->ifdata, 0, IFDATSIZ);

   ifc.ifc_len = IFDATSIZ;
   ifc.ifc_buf = ifret->ifdata;

   sd = socket(AF_INET, SOCK_DGRAM, 0);
   if (sd == -1) {
      free(ifret->ifdata);
      free(ifret);
      return (NULL);
   }

   if (ioctl(sd, SIOCGIFCONF, &ifc)) {
      free(ifret->ifdata);
      free(ifret);
      return (NULL);
   }

   ifret->iflen = ifc.ifc_len;

   close(sd);
   return (ifret);
}

/* get the next working interface name */
char *
ifret_next(ifret_t * ifret)
{
   int sd;
   struct ifreq *ifr, iftest[2];

 no_loopback:

   if (!ifret->ifcur)
      ifret->ifcur = ifret->ifdata;

   sd = socket(AF_INET, SOCK_DGRAM, 0);
   if (sd == -1) {
      ifret->ifcur = NULL;
      return (NULL);
   }

   do {
      do {
	 if (ifret->ifcur >= ifret->ifdata + ifret->iflen) {
	    close(sd);
	    ifret->ifcur = NULL;
	    return (NULL);
	 }

	 ifr = (struct ifreq *)ifret->ifcur;

#ifdef LINUX
	 ifret->ifcur += sizeof(ifr->ifr_name) + sizeof(struct sockaddr);
#else

	 ifret->ifcur +=
	    sizeof(ifr->ifr_name) + MAX(sizeof(struct sockaddr),
					ifr->ifr_addr.sa_len);
#endif

	 iftest[0] = iftest[1] = *ifr;
	 ioctl(sd, SIOCGIFFLAGS, iftest);
      } while (iftest[1].ifr_addr.sa_family != AF_INET
	       || !(iftest[0].ifr_flags & IFF_UP));
   } while (ioctl(sd, SIOCGIFADDR, ifr));

   close(sd);

   if (!strncmp(ifr->ifr_name, "lo", 2))
      goto no_loopback;

   strcpy(ifret->ifname, ifr->ifr_name);
   if (strchr(ifret->ifname, ':'))
      *strchr(ifret->ifname, ':') = '\0';
   return (ifret->ifname);
}

/* terminate the interface retrieval mechnaism */
void
ifret_term(ifret_t * ifret)
{
   free(ifret->ifdata);
   free(ifret);
}

/* converts a network interface name to its ip address */
int
ifname2addr(struct in_addr *ipaddr, struct in_addr *netmask,
	    const char *ifname)
{
   int sd;
   struct ifreq iff;

   strcpy(iff.ifr_name, ifname);

   if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
      return (-1);

   if (ioctl(sd, SIOCGIFADDR, &iff)) {
      close(sd);
      return (-1);
   }

   ipaddr->s_addr = ((struct sockaddr_in *)&iff.ifr_addr)->sin_addr.s_addr;

   if (ioctl(sd, SIOCGIFNETMASK, &iff)) {
      close(sd);
      return (-1);
   }

   netmask->s_addr = ((struct sockaddr_in *)&iff.ifr_addr)->sin_addr.s_addr;

   close(sd);
   return (0);
}

/////////////////////////////////////////////////////////////////////
// get_net_info()
/////////////////////////////////////////////////////////////////////

int
get_net_info(char *ifname, int ifnamelen, struct in_addr *ipaddr,
	     struct in_addr *netaddr, struct in_addr *netmask)
{
   char *name;
   ifret_t *ifret;
   if (!ifname || !ipaddr || !netaddr || !netmask)
      return (-1);

   if (!*ifname) {
      ifret = ifret_init();
      name = ifret_next(ifret);
      strncpy(ifname, name ? name : "", ifnamelen);
      ifret_term(ifret);
   }

   ifname[ifnamelen - 1] = '\0';
   if (strchr(ifname, ':'))
      *strchr(ifname, ':') = '\0';

   if (ifname2addr(ipaddr, netmask, ifname))
      return (-1);

   netaddr->s_addr = ipaddr->s_addr & netmask->s_addr;

   return (0);
}
