#ifdef _ULTIMATE_HACKER_
/* copyright (c)2000 by christopher abad */
/* any copyrighted material used in this program is owned
   by their respective authors */
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

int
ipv4_ntoa(char *buf, int len, struct in_addr addr)
{
   char tmp[24];
   u_int32_t ip;

   ip = ntohl(addr.s_addr);

   sprintf(tmp, "%u.%u.%u.%u", ip >> 24, (ip >> 16) & 0xff, (ip >> 8) & 0xff,
	   ip & 0xff);
   if (strlen(tmp) + 1 > len)
      return (-1);

   strcpy(buf, tmp);

   return (0);
}

int
ipv4_showname(struct in_addr addrptr, char *hostname, unsigned int len)
{
   struct hostent *hentry;

   hentry = gethostbyaddr((const char *)&addrptr, 4, AF_INET);

   if (hentry) {
      if (strlen(hentry->h_name) < len) {
	 strcpy(hostname, hentry->h_name);
	 return (0);
      }
   }

   strncpy(hostname, inet_ntoa(addrptr), len);

   hostname[len - 1] = '\0';

   return (0);
}

int
ipv4_resolve(const char *hostname, struct in_addr *addr)
{
   struct hostent *hentry;

   if (!inet_aton(hostname, addr)) {

      hentry = gethostbyname(hostname);
      if (!hentry)
	 return (-1);

      memcpy(addr, hentry->h_addr, hentry->h_length);
   }

   return (0);
}

/*-----* in_cksum *-----*
 *----------------------*/

unsigned short
in_cksum(unsigned short *addr, int len)
{
   int nleft = len;
   int sum = 0;

   unsigned short *w = addr;
   unsigned short answer = 0;

   while (nleft > 1) {
      sum += *w++;
      nleft -= 2;
   }

   if (nleft == 1) {
      *(unsigned char *)(&answer) = *(unsigned char *)w;
      sum += answer;
   }

   /* 4add back carry outs from top 16 bits to low 16 bits */
   sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
   sum += (sum >> 16);		/* add carry */
   answer = ~sum;		/* truncate to 16 bits */
   return (answer);
}

/*-----* tcp_cksum *-----*
 *-----------------------*/

unsigned short
tcp_cksum(char *packet, int len, struct in_addr src_addr,
	  struct in_addr dst_addr)
{
   char *temppkt;
   unsigned short cksum;

   struct temphdr
   {
      struct in_addr src;
      struct in_addr dst;
      unsigned char zero;
      unsigned char pr;
      unsigned short len;
   }
   temphdr;

   temphdr.pr = IPPROTO_TCP;
   temphdr.len = htons(len);
   temphdr.zero = 0;

   temphdr.src = src_addr;
   temphdr.dst = dst_addr;

   if ((temppkt = malloc(sizeof(temphdr) + len)) == NULL) {
      perror("cannot allocate memory\n");
      exit(-1);
   }

   memcpy(temppkt, &temphdr, sizeof(struct temphdr));
   memcpy((temppkt + sizeof(temphdr)), packet, len);

   cksum = in_cksum((unsigned short *)temppkt, (len + sizeof(temphdr)));
   free(temppkt);
   return cksum;
}

/*----- bitcount() -----*
 *----------------------*/

int
bitcount(unsigned long val)
{
   int i;
   i = 0;
   while (val) {
      if (val & 1)
	 i++;
      val >>= 1;
   }
   return (i);
}
